#!/usr/bin/env python
#
# Copyright (C) 2014 eNovance SAS <licensing@enovance.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import json
try:
    from urllib.parse import urlparse
except Exception:
    # PY2 compat
    from urlparse import urlparse
import requests


class IntrospectionNotAvailableError(Exception):
    pass


def get_cookie(auth_server,
               username=None, password=None,
               github_access_token=None,
               api_key=None,
               use_ssl=True,
               verify=True):
    # TODO: remove this parameter once
    #       I8df68b7f74344371e4b45b4a6d1cc3362b70b61e is merged
    if use_ssl is False:
        use_ssl = True
    if urlparse(auth_server).scheme == '':
        auth_server = "https://%s" % auth_server
    cauth_info = get_cauth_info(auth_server, verify)
    url = "%s/auth/login" % auth_server
    auth_params = {'back': '/',
                   'args': {}, }
    methods = cauth_info['service']['auth_methods']
    if (username and password and ('Password' in methods)):
        auth_params['args'] = {'username': username,
                               'password': password}
        auth_params['method'] = 'Password'
    elif (github_access_token and
          ('GithubPersonalAccessToken' in methods)):
        auth_params['args'] = {'token': github_access_token}
        auth_params['method'] = 'GithubPersonalAccessToken'
    elif (api_key and ('APIKey' in methods)):
        auth_params['args'] = {'api_key': api_key}
        auth_params['method'] = 'APIKey'
    else:
        m = "Missing credentials (accepted auth methods: %s)"
        methods = ','.join(methods)
        raise ValueError(m % methods)
    header = {'Content-Type': 'application/json'}
    resp = requests.post(url, json.dumps(auth_params, sort_keys=True),
                         allow_redirects=False,
                         verify=verify,
                         headers=header)
    return resp.cookies.get('auth_pubtkt', '')


def _get_service_info(url, verify=True):
    resp = requests.get(url, allow_redirects=False,
                        verify=verify)
    return resp.json()


def get_cauth_info(auth_server, verify=True):
    url = "%s/auth/about/" % auth_server
    return _get_service_info(url, verify)


def get_managesf_info(auth_server, verify=True):
    url = "%s/about/" % auth_server
    return _get_service_info(url, verify)



# -*- coding: utf-8 -*-
#
# Copyright (c) 2016 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os

from unittest import TestCase

from managesf.controllers import SFuser
from managesf.tests import dummy_conf


class SFuserController(TestCase):
    @classmethod
    def setupClass(cls):
        cls.conf = dummy_conf()
        SFuser.model.conf = cls.conf

    def setUp(self):
        SFuser.model.init_model()

    def tearDown(self):
        os.unlink(self.conf.sqlalchemy['url'][len('sqlite:///'):])

    def test_create(self):
        u = SFuser.SFUserManager()
        id = u.create(username=u'SpongeBob',
                      email='SquarePants',
                      fullname=u'Sp. Sq.',
                      cauth_id=17)
        self.assertEqual(str(id),
                         SFuser.crud.get(username=u'SpongeBob').get('id'))
        # recreating the same user returns the same id
        id1 = u.create(username=u'SpongeBob',
                       email='SquarePants',
                       fullname=u'Sp. Sq.',
                       cauth_id=17)
        self.assertEqual(str(id1),
                         str(id))
        # if cauth_id is reset, it is updated
        id2 = u.create(username=u'SpongeBob',
                       email='SquarePants',
                       fullname=u'Sp. Sq.',
                       cauth_id=24)
        self.assertEqual(str(id2),
                         str(id))
        self.assertEqual('24',
                         SFuser.crud.get(id2).get('cauth_id'))
        # cauth_id acts as Authoritah
        id3 = u.create(username=u'Jake',
                       email='the Dog',
                       fullname=u'J. the Dog',
                       cauth_id=24)
        self.assertEqual(str(id3),
                         str(id))
        self.assertEqual('Jake',
                         SFuser.crud.get(id3).get('username'))
        self.assertEqual('the Dog',
                         SFuser.crud.get(id3).get('email'))
        self.assertEqual('J. the Dog',
                         SFuser.crud.get(id3).get('fullname'))
        # create another user, default cauth_id is -1
        id4 = u.create(username=u'Finn',
                       email='the Human',
                       fullname=u'F. The Human')
        self.assertTrue(str(id4) != str(id))
        self.assertEqual('-1',
                         SFuser.crud.get(id4).get('cauth_id'))
        # create another user, make sure it does not update Finn
        id5 = u.create(username=u'Bonnibel',
                       email='Bubblegum',
                       fullname=u'Princess Bubblegum')
        self.assertTrue(str(id4) != str(id5))

    def test_get(self):
        u = SFuser.SFUserManager()
        u.create(username=u'Bonnibel',
                 email='Bubblegum',
                 fullname=u'Princess Bubblegum')
        u.create(username=u'Finn',
                 email='the Human',
                 fullname=u'F. The Human')
        id = u.create(username=u'Jake',
                      email='the Dog',
                      fullname=u'J. the Dog',
                      cauth_id=24)
        self.assertEqual('Bonnibel',
                         u.get(username=u'Bonnibel')['username'])
        self.assertEqual('Bubblegum',
                         u.get(username=u'Bonnibel',
                               fullname=u'Princess Bubblegum')['email'])
        self.assertEqual('Finn',
                         u.get(email='the Human')['username'])
        self.assertEqual('the Human',
                         u.get(email='the Human')['email'])
        self.assertEqual('Jake',
                         u.get(fullname=u'J. the Dog')['username'])
        self.assertEqual('the Dog',
                         u.get(fullname=u'J. the Dog')['email'])
        self.assertEqual('Jake',
                         u.get(cauth_id=24)['username'])
        self.assertEqual('Jake',
                         u.get(id)['username'])
        self.assertEqual({},
                         u.get(username=u'BMO'))
        self.assertRaises(KeyError,
                          u.get, cauth_id=-1)

    def test_all(self):
        u = SFuser.SFUserManager()
        u.create(username=u'Bonnibel',
                 email='Bubblegum',
                 fullname=u'Princess Bubblegum')
        u.create(username=u'Finn',
                 email='the Human',
                 fullname=u'F. The Human')
        u.create(username=u'Jake',
                 email='the Dog',
                 fullname=u'J. the Dog',
                 cauth_id=24)
        u.create(username=u'Gunther',
                 email='the Penguin',
                 fullname=u'Orgalorg the Destroyer')
        self.assertEqual(4,
                         len(u.all()))

    def test_update(self):
        u = SFuser.SFUserManager()
        pb = u.create(username=u'Bonnibel',
                      email='Bubblegum',
                      fullname=u'Princess Bubblegum')
        bonnibel = u.get(pb)
        # updating a non existing user does nothing
        self.assertEqual(None,
                         u.update(pb + 2))
        u.update(pb, username=u'Marceline')
        self.assertEqual('Marceline', u.get(pb)['username'])
        u.update(pb, email='The Vampire')
        self.assertEqual('The Vampire', u.get(pb)['email'])
        u.update(pb, fullname=u'M the V')
        self.assertEqual('M the V', u.get(pb)['fullname'])
        u.update(pb, username=u'Bonnibel',
                 email='Bubblegum', fullname=u'Princess Bubblegum')
        self.assertEqual(bonnibel,
                         u.get(pb))

    def test_reset_cauth_id(self):
        u = SFuser.SFUserManager()
        pb = u.create(username=u'Bonnibel',
                      email='Bubblegum',
                      fullname=u'Princess Bubblegum',
                      cauth_id=23)
        bonnibel = u.get(pb)
        self.assertEqual('23',
                         bonnibel['cauth_id'])
        u.reset_cauth_id(pb, 42)
        bonnibel = u.get(pb)
        self.assertEqual('42',
                         bonnibel['cauth_id'])

    def test_delete(self):
        u = SFuser.SFUserManager()
        pb = u.create(username=u'Bonnibel',
                      email='Bubblegum',
                      fullname=u'Princess Bubblegum',
                      cauth_id=23)
        u.delete(pb)
        self.assertEqual({},
                         u.get(username=u'Bonnibel'))
        pb = u.create(username=u'Bonnibel',
                      email='Bubblegum',
                      fullname=u'Princess Bubblegum',
                      cauth_id=23)
        u.delete(username=u'Bonnibel')
        self.assertEqual({},
                         u.get(username=u'Bonnibel'))
        pb = u.create(username=u'Bonnibel',
                      email='Bubblegum',
                      fullname=u'Princess Bubblegum',
                      cauth_id=23)
        u.delete(email='Bubblegum')
        self.assertEqual({},
                         u.get(username=u'Bonnibel'))
        pb = u.create(username=u'Bonnibel',
                      email='Bubblegum',
                      fullname=u'Princess Bubblegum',
                      cauth_id=23)
        u.delete(fullname=u'Princess Bubblegum')
        self.assertEqual({},
                         u.get(username=u'Bonnibel'))
        pb = u.create(username=u'Bonnibel',
                      email='Bubblegum',
                      fullname=u'Princess Bubblegum',
                      cauth_id=23)
        u.delete(cauth_id=23)
        self.assertEqual({},
                         u.get(username=u'Bonnibel'))
        pb = u.create(username=u'Bonnibel',
                      email='Bubblegum',
                      fullname=u'Princess Bubblegum',
                      cauth_id=23)
        self.assertFalse(u.delete(username=u'SusanStrong'))
        u.delete(cauth_id=23, email='Bubblegum')
        self.assertEqual({},
                         u.get(username=u'Bonnibel'))

    def test_unicode(self):
        """create and get a non ascii user"""
        u = SFuser.SFUserManager()
        id = u.create(username=u'七代目火影4lyf',
                      email='datte@bayo',
                      fullname=u'うずまきナルト',
                      cauth_id=999)
        self.assertEqual(u'うずまきナルト',
                         SFuser.crud.get(id=id).get('fullname'))
        by_username = SFuser.crud.get(username=u'七代目火影4lyf')
        self.assertEqual(u'うずまきナルト',
                         by_username.get('fullname'))



#
# Copyright (C) 2014 eNovance SAS <licensing@enovance.com>
# Copyright (C) 2016-2021 Red Hat
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
# -*- coding: utf-8 -*-

import setuptools

setuptools.setup(
    name="managesf",
    setup_requires=['pbr'],
    packages=["managesf"],
    description="REST API interface and CLI for software-factory project",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    pbr=True)



# Copyright (C) 2014 eNovance SAS <licensing@enovance.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import base64
import logging
from urllib.parse import unquote

from pecan import conf
from pecan import expose
from pecan import abort
from pecan.rest import RestController
from pecan import request, response
from stevedore import driver

from managesf.controllers import localuser, introspection
from managesf.controllers import SFuser
from managesf.services import base
from managesf.services import exceptions
from managesf import policy

from managesf import DEFAULT_SERVICES
from managesf.controllers.api.v2 import resources as v2_resources
from managesf.controllers.api.v2 import configurations as v2_configurations

from managesf.api.v2.managers import resource_manager


logger = logging.getLogger(__name__)

LOGERRORMSG = "Unable to process client request, failed with "\
              "unhandled error: %s"
CLIENTERRORMSG = "Unable to process your request, failed with "\
                 "unhandled error (server side): %s"

# TODO: add detail (detail arg or abort function) for all abort calls.


# instanciate service plugins
SF_SERVICES = []
SERVICES = {}


def load_services():
    try:
        if conf.services:
            services = conf.services
        else:
            services = DEFAULT_SERVICES
            msg = 'No service configured, loading: %s' % DEFAULT_SERVICES
            logger.info(msg)
    except AttributeError:
        services = DEFAULT_SERVICES
        msg = 'Obsolete conf file, loading default: %s' % DEFAULT_SERVICES
        logger.info(msg)

    for service in services:
        try:
            plugin = driver.DriverManager(namespace='managesf.service',
                                          name=service,
                                          invoke_on_load=True,
                                          invoke_args=(conf,)).driver
            SF_SERVICES.append(plugin)
            SERVICES[service] = plugin
            logger.info('%s plugin loaded successfully' % service)
        except Exception as e:
            logger.error('Could not load service %s: %s' % (service, e))


def _decode_project_name(name):
    if name.startswith('==='):
        try:
            n = base64.urlsafe_b64decode(name.encode()[3:])
            return n.decode('utf8')
        except Exception:
            return name[3:]
    return name


class SFManager:
    user = SFuser.SFUserManager()


sfmanager = SFManager()


def is_admin(user):
    return base.RoleManager.is_admin(user)


def report_unhandled_error(exp):
    logger.exception(LOGERRORMSG % str(exp))
    response.status = 500
    return CLIENTERRORMSG % str(exp)

# TODO move to utils and use resources rather than gerrit for groups


def authorize(rule_name, target):
    if not request.remote_user:
        request.remote_user = request.headers.get('X-Remote-User')
    credentials = {'username': request.remote_user, 'groups': []}
    # OpenID Connect authentication
    if request.headers.get("OIDC_CLAIM_groups", None) is not None:
        for group in request.headers.get('OIDC_CLAIM_groups').split(','):
            # it seems like keycloak prefixes groups with /, remove it
            if group.startswith('/'):
                credentials['groups'].append(group[1:])
            else:
                credentials['groups'].append(group)
    # gerrit based
    else:
        if request.remote_user:
            code_reviews = [s for s in SF_SERVICES
                            if isinstance(s, base.BaseCodeReviewServicePlugin)]
            if code_reviews:
                user_groups = code_reviews[0].project.get_user_groups(
                    request.remote_user)
                credentials['groups'] = [grp['name'] for grp in user_groups]
    return policy.authorize(rule_name, target, credentials)


class LocalUserController(RestController):

    @expose('json')
    def post(self, username):
        _policy = 'managesf.localuser:create_update'
        if not authorize(_policy,
                         target={'username': username}):
            return abort(401,
                         detail='Failure to comply with policy %s' % _policy)
        infos = request.json if request.content_length else {}
        try:
            ret = localuser.update_user(username, infos)
        except (localuser.InvalidInfosInput, localuser.BadUserInfos) as e:
            abort(400, detail=str(e))
        except Exception as e:
            return report_unhandled_error(e)
        if isinstance(ret, dict):
            # user created - set correct status code
            response.status = 201
        return ret

    @expose('json')
    def get(self, username):
        _policy = 'managesf.localuser:get'
        if not authorize(_policy,
                         target={'username': username}):
            return abort(401,
                         detail='Failure to comply with policy %s' % _policy)
        try:
            ret = localuser.get_user(username)
        except localuser.UserNotFound as e:
            abort(404, detail=str(e))
        except Exception as e:
            return report_unhandled_error(e)
        return ret

    @expose('json')
    def delete(self, username):
        _policy = 'managesf.localuser:delete'
        if not authorize(_policy,
                         target={'username': username}):
            return abort(401,
                         detail='Failure to comply with policy %s' % _policy)
        try:
            ret = localuser.delete_user(username)
        except localuser.UserNotFound as e:
            abort(404, detail=str(e))
        except Exception as e:
            return report_unhandled_error(e)
        return ret


class LocalUserBindController(RestController):

    log = logging.getLogger("BindController")

    @expose('json')
    def get(self):
        _policy = 'managesf.localuser:bind'
        authorization = request.headers.get('Authorization', None)
        if not authorization:
            abort(401, detail="Authentication header missing")
        try:
            username, password = localuser.decode(authorization)
        except localuser.DecodeError:
            self.log.warning("Authorization decoding error")
            abort(401, detail="Wrong authorization header")
        if not authorize(_policy,
                         target={'username': username}):
            self.log.error(u"%s: policy error" % username)
            return abort(401,
                         detail='Failure to comply with policy %s' % _policy)

        try:
            ret = localuser.bind_user(authorization)
        except (localuser.BindForbidden, localuser.UserNotFound) as e:
            self.log.warning(u"%s: UserNotFound or Forbidden" % username)
            abort(401, detail=str(e))
        except Exception as e:
            self.log.exception(u"%s: couldn't bind user" % username)
            return report_unhandled_error(e)
        if not ret:
            self.log.exception(u"%s: Authentication failed" % username)
            abort(401, detail="Authentication failed")
        self.log.info(u"%s: binding success" % username)
        return ret


class ServicesUsersController(RestController):

    def _remove_non_updatable_fields(self, infos):
        forbidden = sum([s.user.check_forbidden_fields(**infos)
                         for s in SF_SERVICES], [])
        msg = 'The following fields cannot be updated: %s, discarding them'
        logger.debug(msg % str(forbidden))
        return dict((u, infos[u]) for u in infos.keys()
                    if u not in forbidden and infos[u] is not None)

    def _update(self, user_id, infos):
        sfmanager.user.update(user_id,
                              username=infos.get('username'),
                              email=infos.get('email'),
                              fullname=infos.get('full_name'),
                              idp_sync=infos.get('idp_sync'))
        for service in SF_SERVICES:
            s_id = sfmanager.user.mapping.get_service_mapping(
                service.service_name,
                user_id)
            if s_id:
                try:
                    service.user.update(uid=s_id, **infos)
                except exceptions.UnavailableActionError:
                    pass
            else:
                full_name = infos.get('full_name')
                username = infos.get('username')
                ssh_keys = infos.get('ssh_keys', [])
                email = infos.get('email')
                cauth_id = infos.get('external_id')
                try:
                    s_id = service.user.create(username=username,
                                               email=email,
                                               full_name=full_name,
                                               ssh_keys=ssh_keys,
                                               cauth_id=cauth_id)
                    sfmanager.user.mapping.set(user_id,
                                               service.service_name,
                                               s_id)
                except exceptions.UnavailableActionError:
                    pass

    @expose('json')
    def put(self, id=None, email=None, username=None):
        _policy = 'managesf.user:update'
        infos = request.json if request.content_length else {}
        # the JSON payload is only for data to update.
        _email = request.GET.get('email')
        if _email:
            _email = unquote(_email)
        _username = request.GET.get('username')
        if _username:
            _username = unquote(_username)
        d_id = request.GET.get('id')
        if not d_id and (_email or _username):
            logger.debug('[update] looking for %s %s ...' % (_email,
                                                             _username))
            d_id = sfmanager.user.get(username=_username,
                                      email=_email).get('id')
            logger.debug('found %s %s with id %s' % (_email, _username, d_id))
        if not d_id:
            response.status = 404
            return
        u = _username or sfmanager.user.get(d_id).get('username')
        if not authorize(_policy,
                         target={'username': u}):
            return abort(401,
                         detail='Failure to comply with policy %s' % _policy)
        sanitized = self._remove_non_updatable_fields(infos)
        logger.debug('[update] sanitized request %r to %r' % (infos,
                                                              sanitized))
        if not sanitized:
            if sanitized != infos:
                msg = 'You tried to update immutable fields'
            else:
                msg = 'Nothing to do'
            abort(400,
                  detail=msg)
        try:
            self._update(d_id, sanitized)
            response.status = 200
            return {'updated_fields': sanitized}
        except Exception as e:
            return report_unhandled_error(e)

    @expose('json')
    def post(self):
        _policy = 'managesf.user:create'
        infos = request.json if request.content_length else {}
        if not infos or not infos.get('username'):
            abort(400, detail=u'Incomplete user information: %r' % infos)
        if not authorize(_policy,
                         target={'username': infos.get('username')}):
            return abort(401,
                         detail='Failure to comply with policy %s' % _policy)
        try:
            known_user = None
            if infos.get('external_id', -1) != -1:
                known_user = sfmanager.user.get(cauth_id=infos['external_id'])
                if known_user:
                    msg = (u'found user #%(id)s %(username)s (%(email)s) '
                           u'by cauth ID #%(cauth_id)s, user needs update')
                    logger.debug(msg % known_user)
                    u = known_user['id']
                    clean_infos = self._remove_non_updatable_fields(infos)
                    if known_user.get('idp_sync'):
                        self._update(u, clean_infos)
                    else:
                        logger.info("Skipping user information update because"
                                    "idp_sync is disabled")
            # if we still cannot find it, let's create it
            if not known_user:
                u = sfmanager.user.create(username=infos['username'],
                                          email=infos['email'],
                                          fullname=infos['full_name'],
                                          cauth_id=infos.get('external_id'))
                self._create_user_in_services(u, infos)
        except Exception as e:
            return report_unhandled_error(e)
        # TODO(mhu) later, this should return the local id and the user data
        response.status = 201

    def _create_user_in_services(self, user_id, infos):
        for service in SF_SERVICES:
            try:
                s_id = service.user.get(username=infos.get('username'))
                if s_id:
                    msg = u'[%s] user %s exists, skipping creation'
                    logger.debug(msg % (service.service_name,
                                        infos.get('username')))
                    mapped = sfmanager.user.mapping.get_user_mapping(
                        service.service_name,
                        s_id)
                    if not mapped:
                        sfmanager.user.mapping.set(user_id,
                                                   service.service_name,
                                                   s_id)
                        msg = u'[%s] user %s mapped to id %s'
                        logger.debug(msg % (service.service_name,
                                            infos.get('username'),
                                            s_id))
                else:
                    full_name = infos.get('full_name')
                    username = infos.get('username')
                    ssh_keys = infos.get('ssh_keys', [])
                    email = infos.get('email')
                    cauth_id = infos.get('external_id')
                    s_id = service.user.create(username=username,
                                               email=email,
                                               full_name=full_name,
                                               ssh_keys=ssh_keys,
                                               cauth_id=cauth_id)
                    # we might have a mapping, but to a wrong user id in the
                    # service (because the user existed before but was removed
                    # directly from the service, for example)
                    mapped = sfmanager.user.mapping.get_service_mapping(
                        service.service_name,
                        user_id)
                    if mapped and mapped != s_id:
                        msg = u'[%s] user %s wrongly mapped to id %s, removing'
                        logger.debug(msg % (service.service_name,
                                            infos.get('username'),
                                            mapped))
                        sfmanager.user.mapping.delete(user_id,
                                                      service.service_name,
                                                      mapped)
                    sfmanager.user.mapping.set(user_id,
                                               service.service_name,
                                               s_id)
                    msg = u'[%s] user %s mapped to %s id %s'
                    logger.debug(msg % (service.service_name,
                                        infos.get('username'),
                                        service.service_name,
                                        s_id))
            except exceptions.UnavailableActionError:
                pass

    @expose('json')
    def get(self, **kwargs):
        _policy = 'managesf.user:get'
        if not authorize(_policy,
                         target={'username': kwargs.get('username')}):
            return abort(401,
                         detail='Failure to comply with policy %s' % _policy)
        return sfmanager.user.get(**kwargs)

    @expose()
    def delete(self, id=None, email=None, username=None):
        _policy = 'managesf.user:delete'
        if not authorize(_policy,
                         target={'username': username}):
            return abort(401,
                         detail='Failure to comply with policy %s' % _policy)
        d_id = id
        if not d_id and (email or username):
            logger.debug(u'[delete] looking for %s %s' % (email, username))
            d_id = sfmanager.user.get(username=username,
                                      email=email).get('id')
        if not d_id:
            response.status = 404
            return
        logger.debug(u'found %s %s with id %s' % (email, username, d_id))
        try:
            for service in SF_SERVICES:
                try:
                    service.user.delete(username=username)
                    sfmanager.user.mapping.delete(d_id,
                                                  service.service_name)
                except exceptions.UnavailableActionError:
                    pass
            sfmanager.user.delete(id=d_id)
        except Exception as e:
            return report_unhandled_error(e)
        response.status = 204


class HooksController(RestController):

    def get_project_by_repo(self, reponame):
        for _, project in resource_manager.resources.get().get(
                'resources', {}).get('projects', {}).items():
            for repo in project.get('source-repositories', []):
                if reponame in repo:
                    return project
        return None

    @expose('json')
    def post(self, hook_name):
        """Trigger hook {hook_name}."""
        _policy = 'managesf.hooks:trigger'
        if not authorize(_policy,
                         target={}):
            return abort(401,
                         detail='Failure to comply with policy %s' % _policy)
        change = request.json if request.content_length else {}
        # Get hook engine configuration for project
        project_name = change.get('project')
        if not project_name:
            return abort(400, detail=u"Hooks: Invalid change %s" % change)
        project = self.get_project_by_repo(project_name)
        if not project:
            logger.info("Hooks: Repository %s is not part of any project" %
                        change.get('project'))
            return abort(204, detail=u"No issue-tracker defined")
        hook = project.get('issue-tracker')

        status = 200
        try:
            msg = getattr(SERVICES[hook].hooks, hook_name)(**change)
        except Exception as e:
            status = 400
            msg = str(e)
            logger.error(u"[%s] hook %s failed with %s" % (
                hook, hook_name, msg))
        response.status = status
        return {'msg': msg}


load_services()


# API v2 - will be in its own file hierarchy once dependencies are sorted out


class V2Controller(object):
    # Mimic api v1 and replace endpoints incrementally
    user = LocalUserController()
    bind = LocalUserBindController()
    about = introspection.IntrospectionController()
    services_users = ServicesUsersController()
    hooks = HooksController()
    resources = v2_resources.ResourcesRootController()
    configurations = v2_configurations.ConfigurationController()


class RootController(object):
    def __init__(self, *args, **kwargs):
        try:
            # just try to get the api config
            _ = conf.api.v2  # noQA
            self.v2 = V2Controller()
        except AttributeError:
            # TODO have a generic blank REST controller that returns
            # 'Not Implemented' error code
            logger.info('API v2 is not configured, skipping endpoint.')
            self.v2 = RestController()

        self.user = LocalUserController()
        self.bind = LocalUserBindController()
        self.about = introspection.IntrospectionController()
        self.services_users = ServicesUsersController()
        self.hooks = HooksController()



# Copyright (C) 2014 eNovance SAS <licensing@enovance.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os
import shutil

from unittest import TestCase
from webtest import TestApp
from pecan import load_app
from mock import patch, MagicMock

from basicauth import encode

from managesf.tests import dummy_conf

from managesf.services import exceptions as exc

# plugins imports
# TODO: should be done dynamically depending on what plugins we want

from managesf.services.gerrit.project import SFGerritProjectManager
from managesf.services.gerrit import user as g_user


def raiseexc(*args, **kwargs):
    raise Exception('FakeExcMsg')


class FunctionalTest(TestCase):
    def setUp(self):
        self.to_delete = []
        c = dummy_conf()
        self.config = {'services': c.services,
                       'gerrit': c.gerrit,
                       'app': c.app,
                       'admin': c.admin,
                       'sqlalchemy': c.sqlalchemy,
                       'managesf': c.managesf,
                       'policy': c.policy,
                       'resources': c.resources,
                       'api': c.api, }
        # deactivate loggin that polute test output
        # even nologcapture option of nose effetcs
        # 'logging': c.logging}
        self.app = TestApp(load_app(self.config))

    def tearDown(self):
        # Remove the sqlite db
        os.unlink(self.config['sqlalchemy']['url'][len('sqlite:///'):])
        for path in self.to_delete:
            if os.path.isdir(path):
                shutil.rmtree(path)
            else:
                os.unlink(path)


class TestManageSFAPIv2(FunctionalTest):
    def test_get_api_endpoint(self):
        response = self.app.get('/v2/about/').json
        self.assertEqual('managesf',
                         response['service']['name'])
        self.assertEqual(set(self.config['services']),
                         set(response['service']['services']))


class TestManageSFIntrospectionController(FunctionalTest):

    def test_instrospection(self):
        response = self.app.get('/about/').json
        self.assertEqual('managesf',
                         response['service']['name'])
        self.assertEqual(set(self.config['services']),
                         set(response['service']['services']))


class TestManageSFAppLocaluserController(FunctionalTest):

    def test_add_or_update_user(self):
        with patch.object(SFGerritProjectManager, 'get_user_groups') as gug:
            gug.return_value = []
            environ = {'REMOTE_USER': 'admin'}
            infos = {'email': 'john@tests.dom', 'sshkey': 'sshkey',
                     'fullname': 'John Doe', 'password': 'secret'}
            response = self.app.post_json('/user/john', infos,
                                          extra_environ=environ, status="*")
            self.assertEqual(response.status_int, 201)
            infos = {'email': 'john2@tests.dom', 'sshkey': 'sshkey',
                     'fullname': 'John Doe', 'password': 'bigsecret'}
            response = self.app.post_json('/user/john', infos,
                                          extra_environ=environ, status="*")
            self.assertEqual(response.status_int, 200)
            infos = {'wrongkey': 'heyhey'}
            response = self.app.post_json('/user/john', infos,
                                          extra_environ=environ, status="*")
            self.assertEqual(response.status_int, 400)

            # Only admin can add user to that database
            environ = {'REMOTE_USER': 'boss'}
            infos = {'email': 'john2@tests.dom', 'sshkey': 'sshkey',
                     'fullname': 'John Doe 2', 'password': 'secret'}
            response = self.app.post_json('/user/john2', infos,
                                          extra_environ=environ, status="*")
            self.assertEqual(response.status_int, 401)

            # john can update his own data
            environ = {'REMOTE_USER': 'john'}
            infos = {'email': 'otherjohn@tests.dom', 'sshkey': 'sshkey',
                     'fullname': 'other john', 'password': 'secret'}
            response = self.app.post_json('/user/john', infos,
                                          extra_environ=environ, status="*")
            self.assertEqual(response.status_int, 200)
            response = self.app.get('/user/john',
                                    extra_environ=environ, status="*")
            expected = {u'sshkey': u'sshkey',
                        u'username': u'john',
                        u'email': u'otherjohn@tests.dom',
                        u'fullname': u'other john'}
            self.assertEqual(response.status_int, 200)
            self.assertDictEqual(response.json, expected)

            # but jack cannot update john's data
            environ = {'REMOTE_USER': 'jack'}
            infos = {'email': 'jack@tests.dom', 'sshkey': 'sshkey',
                     'fullname': 'handsome jack', 'password': 'secret'}
            response = self.app.post_json('/user/john', infos,
                                          extra_environ=environ, status="*")
            self.assertEqual(response.status_int, 401)
            # jack can read it though
            response = self.app.get('/user/john',
                                    extra_environ=environ, status="*")
            expected = {u'sshkey': u'sshkey',
                        u'username': u'john',
                        u'email': u'otherjohn@tests.dom',
                        u'fullname': u'other john'}
            self.assertEqual(response.status_int, 200)
            self.assertDictEqual(response.json, expected)

    def test_get_user(self):
        with patch.object(SFGerritProjectManager, 'get_user_groups') as gug:
            gug.return_value = []
            environ = {'REMOTE_USER': 'admin'}
            infos = {'email': 'john@tests.dom', 'sshkey': 'sshkey',
                     'fullname': 'John Doe', 'password': 'secret'}
            response = self.app.post_json('/user/john', infos,
                                          extra_environ=environ, status="*")
            self.assertEqual(response.status_int, 201)

            response = self.app.get('/user/john',
                                    extra_environ=environ, status="*")
            expected = {u'sshkey': u'sshkey',
                        u'username': u'john',
                        u'email': u'john@tests.dom',
                        u'fullname': u'John Doe'}
            self.assertEqual(response.status_int, 200)
            self.assertDictEqual(response.json, expected)

            response = self.app.get('/user/notexists',
                                    extra_environ=environ, status="*")
            self.assertEqual(response.status_int, 404)

            environ = {'REMOTE_USER': 'john'}
            response = self.app.get('/user/john',
                                    extra_environ=environ, status="*")
            expected = {u'sshkey': u'sshkey',
                        u'username': u'john',
                        u'email': u'john@tests.dom',
                        u'fullname': u'John Doe'}
            self.assertEqual(response.status_int, 200)
            self.assertDictEqual(response.json, expected)

            environ = {'REMOTE_USER': 'boss'}
            response = self.app.get('/user/john',
                                    extra_environ=environ, status="*")
            expected = {u'sshkey': u'sshkey',
                        u'username': u'john',
                        u'email': u'john@tests.dom',
                        u'fullname': u'John Doe'}
            self.assertEqual(response.status_int, 200)
            self.assertDictEqual(response.json, expected)

    def test_delete_user(self):
        with patch.object(SFGerritProjectManager, 'get_user_groups') as gug:
            gug.return_value = []
            environ = {'REMOTE_USER': 'admin'}
            infos = {'email': 'john@tests.dom', 'sshkey': 'sshkey',
                     'fullname': 'John Doe', 'password': 'secret'}
            response = self.app.post_json('/user/john', infos,
                                          extra_environ=environ, status="*")
            self.assertEqual(response.status_int, 201)

            environ = {'REMOTE_USER': 'boss'}
            response = self.app.delete('/user/john',
                                       extra_environ=environ, status="*")
            self.assertEqual(response.status_int, 401)

            environ = {'REMOTE_USER': 'admin'}
            response = self.app.delete('/user/john',
                                       extra_environ=environ, status="*")
            self.assertEqual(response.status_int, 200)
            response = self.app.get('/user/john',
                                    extra_environ=environ, status="*")
            self.assertEqual(response.status_int, 404)

    def test_bind_user(self):
        with patch.object(SFGerritProjectManager, 'get_user_groups') as gug:
            gug.return_value = []
            environ = {'REMOTE_USER': 'admin'}
            base_infos = {'email': 'john@tests.dom', 'sshkey': 'sshkey',
                          'fullname': 'John Doe', }
            infos = {'password': 'secret'}
            public_infos = {'username': 'john'}
            infos.update(base_infos)
            public_infos.update(base_infos)
            response = self.app.post_json('/user/john', infos,
                                          extra_environ=environ, status="*")
            self.assertEqual(response.status_int, 201)

            auth = encode(u"john", "secret")
            response = self.app.get(
                '/bind', headers={"Authorization": auth}, status="*")
            self.assertEqual(response.status_int, 200)
            self.assertEqual(public_infos,
                             response.json,
                             response.json)

            auth = encode(u"john", "badsecret")
            response = self.app.get(
                '/bind', headers={"Authorization": auth}, status="*")
            self.assertEqual(response.status_int, 401)

            auth = encode(u"boss", "secret")
            response = self.app.get(
                '/bind', headers={"Authorization": auth}, status="*")
            self.assertEqual(response.status_int, 401)


def project_get(*args, **kwargs):
    if kwargs.get('by_user'):
        return ['p1', ]
    return ['p0', 'p1']


class TestManageSFServicesUserController(FunctionalTest):

    def test_add_user_in_backends_non_admin(self):
        environ = {'REMOTE_USER': 'dio'}
        infos = {'email': 'jojo@starplatinum.dom',
                 'ssh_keys': ['ora', 'oraora'],
                 'full_name': 'Jotaro Kujoh', 'username': 'jojo'}
        with patch.object(SFGerritProjectManager, 'get_user_groups') as gug:
            gug.return_value = []
            response = self.app.post_json('/services_users/', infos,
                                          extra_environ=environ, status="*")
            self.assertEqual(response.status_int, 401)

    def test_add_user_in_backends(self):
        environ = {'REMOTE_USER': 'admin'}
        rm_user = MagicMock()
        rm_user.id = 9
        sb_user = MagicMock()
        sb_user.id = 10
        gerrit_created = {"_account_id": 5,
                          "name": "Jotaro Kujoh",
                          "email": "jojo@starplatinum.dom",
                          "username": "jojo",
                          "avatars": [{"url": "meh",
                                       "height": 26}]}
        infos = {'email': 'jojo@starplatinum.dom',
                 'ssh_keys': ['ora', 'oraora'],
                 'full_name': 'Jotaro Kujoh', 'username': 'jojo'}
        rm_user2 = MagicMock()
        rm_user2.id = 12
        sb_user2 = MagicMock()
        sb_user2.id = 13
        gerrit2_created = {"_account_id": 8,
                           "name": "Dio Brando",
                           "email": "dio@wryyyyy.dom",
                           "username": "dio-sama",
                           "avatars": [{"url": "meh",
                                        "height": 26}]}
        infos2 = {'email': 'dio@wryyyyy.dom',
                  'ssh_keys': ['ora', 'oraora'],
                  'full_name': 'Dio Brando', 'username': 'dio-sama'}
        infos3 = {'email': 'john@joestar.dom',
                  'ssh_keys': ['ora', 'oraora'],
                  'full_name': 'Jonathan Joestar', 'username': 'Jon'}
        with patch.object(g_user.SFGerritUserManager,
                          'create') as gerrit_create, \
                patch.object(g_user.SFGerritUserManager, 'get') as g_get, \
                patch.object(SFGerritProjectManager,
                             'get_user_groups'):
            g_get.return_value = None
            gerrit_create.return_value = gerrit_created["_account_id"]
            response = self.app.post_json('/services_users/', infos,
                                          extra_environ=environ, status="*")
            self.assertEqual(response.status_int, 201)
            _, gerrit_args = gerrit_create.call_args
            for k, v in (
                ("username", infos.get('username')),
                ("email", infos.get('email')),
                ("full_name", infos.get('full_name')),
                ("ssh_keys", infos['ssh_keys'])
            ):
                self.assertEqual(gerrit_args[k], v)
            self.assertTrue("cauth_id" in gerrit_args)
            # TODO(mhu) test if mapping is set correctly
        # mock at a lower level
        with patch('managesf.services.gerrit.utils.'
                   'GerritClient.add_pubkey'), \
                patch.object(g_user.SFGerritUserManager,
                             '_add_account_as_external'), \
                patch('managesf.services.gerrit.utils.'
                      'GerritClient.create_account') \
                as create_account, \
                patch.object(g_user.SFGerritUserManager, 'get') as g_get, \
                patch.object(SFGerritProjectManager, 'get_user_groups'):
            create_account.return_value = gerrit2_created
            g_get.return_value = None
            response = self.app.post_json('/services_users/', infos2,
                                          extra_environ=environ, status="*")
            self.assertEqual(response.status_int, 201)

        with patch.object(g_user.SFGerritUserManager,
                          'create') as gerrit_create, \
                patch.object(g_user.SFGerritUserManager, 'get') as g_get, \
                patch.object(SFGerritProjectManager,
                             'get_user_groups'):
            # assert that raising UnavailableActionError won't fail
            def unavailable(*args, **kwargs):
                raise exc.UnavailableActionError
            g_get.side_effect = unavailable
            gerrit_create.return_value = 14
            response = self.app.post_json('/services_users/', infos3,
                                          extra_environ=environ, status="*")
            self.assertEqual(response.status_int, 201)

        with patch.object(g_user.SFGerritUserManager,
                          '_add_account_as_external'), \
                patch('managesf.services.gerrit.utils.'
                      'GerritClient.create_account') \
                as create_account, \
                patch.object(g_user.SFGerritUserManager, 'get') as g_get, \
                patch.object(SFGerritProjectManager, 'get_user_groups'):

            g_get.return_value = gerrit_created["_account_id"]
            create_account.return_value = gerrit_created
            response = self.app.post_json('/services_users/', infos,
                                          extra_environ=environ, status="*")
            self.assertEqual(response.status_int, 201)

        with patch.object(g_user.SFGerritUserManager,
                          '_add_account_as_external'), \
                patch('managesf.services.gerrit.utils.'
                      'GerritClient.create_account') \
                as create_account, \
                patch.object(g_user.SFGerritUserManager, 'get') as g_get, \
                patch.object(SFGerritProjectManager, 'get_user_groups'):
            # assert that user found in backend will skip gracefully
            g_get.return_value = gerrit_created["_account_id"]
            create_account.return_value = gerrit_created
            response = self.app.post_json('/services_users/', infos,
                                          extra_environ=environ, status="*")
            self.assertEqual(response.status_int, 201)

    def test_delete_user_in_backends_non_admin(self):
        environ = {'REMOTE_USER': 'dio'}
        params = {'username': 'iggy'}
        with patch.object(SFGerritProjectManager, 'get_user_groups') as gug:
            gug.return_value = []
            response = self.app.delete('/services_users/', params,
                                       extra_environ=environ, status="*")
            self.assertEqual(response.status_int, 401)

    def test_delete_user_in_backends(self):
        environ = {'REMOTE_USER': 'admin'}
        infos = {'email': 'iggy@stooges.dom',
                 'ssh_keys': ['ora', 'oraora'],
                 'full_name': 'iggy the fool', 'username': 'iggy'}
        with patch.object(g_user.SFGerritUserManager,
                          'delete') as gerrit_delete, \
                patch.object(g_user.SFGerritUserManager,
                             'create') as gerrit_create, \
                patch.object(g_user.SFGerritUserManager,
                             'get') as gerrit_get, \
                patch.object(SFGerritProjectManager, 'get_user_groups') as gug:
            gug.return_value = []
            # Test deletion of non existing user
            response = self.app.delete('/services_users/?username=iggy',
                                       extra_environ=environ, status="*")
            self.assertEqual(response.status_int, 404)
            # test deletion of existing user
            gerrit_create.return_value = 99
            gerrit_get.return_value = None
            self.app.post_json('/services_users/', infos,
                               extra_environ=environ, status="*")
            response = self.app.delete('/services_users/?username=iggy',
                                       extra_environ=environ, status="*")
            self.assertEqual(response.status_int, 204)
            gerrit_delete.assert_called_with(username='iggy')

    def test_all(self):
        environ = {'REMOTE_USER': 'admin'}
        infos_kira = {'email': 'kira@jojolion.dom',
                      'ssh_keys': ['ora', 'oraora'],
                      'full_name': 'yoshikage kira', 'username': 'kira'}
        with patch.object(g_user.SFGerritUserManager, 'delete'), \
                patch.object(g_user.SFGerritUserManager,
                             'create') as g_create, \
                patch.object(g_user.SFGerritUserManager, 'get') as g_get, \
                patch.object(SFGerritProjectManager, 'get_user_groups'):

            g_get.return_value = None
            g_create.return_value = 13
            response = self.app.post_json('/services_users/', infos_kira,
                                          extra_environ=environ, status="*")
            self.assertEqual(response.status_int, 201)
            response = self.app.get('/services_users/',
                                    extra_environ=environ, status="*")
            user_list = response.json
            self.assertTrue(len(user_list) >= 1,
                            user_list)
            self.assertTrue(any(x['username'] == 'kira' for x in user_list))

    def test_get_user(self):
        environ = {'REMOTE_USER': 'admin'}
        infos_jojo = {'email': 'jojo@starplatinum.dom',
                      'ssh_keys': ['ora', 'oraora'],
                      'full_name': 'Jotaro Kujoh', 'username': 'jojo'}
        infos_poln = {'email': 'polnareff@chariot.dom',
                      'ssh_keys': ['ora', 'oraora'],
                      'full_name': 'Polnareff', 'username': 'chariot'}
        with patch.object(g_user.SFGerritUserManager, 'delete'), \
                patch.object(g_user.SFGerritUserManager,
                             'create') as g_create, \
                patch.object(g_user.SFGerritUserManager,
                             'get') as g_get, \
                patch.object(SFGerritProjectManager, 'get_user_groups') as gug:

            g_get.return_value = None
            g_create.return_value = 13
            gug.return_value = []
            response = self.app.post_json('/services_users/', infos_jojo,
                                          extra_environ=environ, status="*")
            self.assertEqual(response.status_int, 201)
            g_create.return_value = 15
            response = self.app.post_json('/services_users/', infos_poln,
                                          extra_environ=environ, status="*")
            self.assertEqual(response.status_int, 201)
            # endpoint allowed to all authenticated users
            environ = {'REMOTE_USER': 'dio'}
            jojo = self.app.get('/services_users/?username=jojo',
                                extra_environ=environ, status="*")
            self.assertEqual(200, jojo.status_int)
            self.assertEqual('Jotaro Kujoh', jojo.json.get('fullname'))
            self.assertTrue('id' in jojo.json.keys())
            self.assertEqual('-1', jojo.json.get('cauth_id'))
            jojo = self.app.get('/services_users/?fullname=Jotaro%20Kujoh',
                                extra_environ=environ, status="*")
            self.assertEqual(200, jojo.status_int)
            self.assertEqual('Jotaro Kujoh', jojo.json.get('fullname'))
            self.assertTrue('id' in jojo.json.keys())
            self.assertEqual('-1', jojo.json.get('cauth_id'))
            # no user
            jojo = self.app.get('/services_users/?username=dio',
                                extra_environ=environ, status="*")
            self.assertEqual(200, jojo.status_int)
            self.assertEqual({}, jojo.json)
            # retrieve all users
            jojo = self.app.get('/services_users/',
                                extra_environ=environ, status="*")
            self.assertEqual(200, jojo.status_int)
            self.assertEqual(2, len(jojo.json))
            # update external id
            infos_jojo = {'email': 'jojo@starplatinum.dom',
                          'ssh_keys': ['ora', 'oraora'],
                          'full_name': 'Jotaro Kujoh', 'username': 'jojo',
                          'external_id': 99}
            environ = {'REMOTE_USER': 'admin'}
            response = self.app.post_json('/services_users/', infos_jojo,
                                          extra_environ=environ, status="*")
            jojo = self.app.get('/services_users/?username=jojo',
                                extra_environ=environ, status="*")
            self.assertEqual('99', jojo.json.get('cauth_id'))

    def test_idp_sync(self):
        environ = {'REMOTE_USER': 'admin'}
        infos_jojo = {'email': 'jojo@starplatinum.dom',
                      'ssh_keys': ['ora', 'oraora'],
                      'full_name': 'Jotaro Kujoh', 'username': 'jojo',
                      'external_id': 42}
        with patch.object(g_user.SFGerritUserManager, 'update') as g_up, \
                patch.object(g_user.SFGerritUserManager,
                             'create') as g_create, \
                patch.object(g_user.SFGerritUserManager,
                             'get') as g_get, \
                patch.object(SFGerritProjectManager, 'get_user_groups'):
            for mock in (g_up, g_get):
                mock.return_value = None
            g_create.return_value = 13

            # User logs in
            response = self.app.post_json('/services_users/', infos_jojo,
                                          extra_environ=environ, status="*")
            self.assertEqual(response.status_int, 201)

            # User disable idp_sync
            disable_idp = {'idp_sync': False}
            user_cred = {'REMOTE_USER': 'jojo'}
            response = self.app.put_json('/services_users/?username=jojo',
                                         disable_idp, extra_environ=user_cred,
                                         status="*")
            self.assertEqual(response.status_int, 200)

            # User change its mail on idp and logs in again
            infos_jojo['email'] = 'jojo@galactica.com'
            response = self.app.post_json('/services_users/', infos_jojo,
                                          extra_environ=environ, status="*")
            self.assertEqual(response.status_int, 201)

            # Check user mail didn't change because idp_sync is False
            response = self.app.get('/services_users/?username=jojo',
                                    extra_environ=environ, status="*")
            self.assertNotEqual(response.json['email'], infos_jojo['email'])

    def test_update_user(self):
        environ = {'REMOTE_USER': 'admin'}
        infos_jojo = {'email': 'jojo@starplatinum.dom',
                      'ssh_keys': ['ora', 'oraora'],
                      'full_name': 'Jotaro Kujoh', 'username': 'jojo'}
        with patch.object(g_user.SFGerritUserManager,
                          'create') as g_create, \
                patch.object(g_user.SFGerritUserManager,
                             'update') as g_update, \
                patch.object(g_user.SFGerritUserManager,
                             'get') as g_get, \
                patch.object(SFGerritProjectManager, 'get_user_groups') as gug:

            gug.return_value = []
            for mock in (g_get, g_update):
                mock.return_value = None
            g_create.return_value = 13
            response = self.app.post_json('/services_users/', infos_jojo,
                                          extra_environ=environ, status="*")
            response = self.app.get('/services_users/?username=jojo',
                                    extra_environ=environ, status="*")
            jojo_id = response.json.get('id')
            # try to update as someone else
            environ = {'REMOTE_USER': 'dio'}
            payload = {'full_name': 'Dio Brando'}
            resp = self.app.put_json('/services_users/?username=jojo',
                                     payload,
                                     extra_environ=environ, status="*")
            self.assertEqual(401, resp.status_int)
            resp = self.app.put_json(
                '/services_users/?email=jojo@starplatinum.dom',
                payload, extra_environ=environ, status="*")
            self.assertEqual(401, resp.status_int)
            resp = self.app.put_json(
                '/services_users/?id=%s' % jojo_id,
                payload, extra_environ=environ, status="*")
            self.assertEqual(401, resp.status_int)
            # try empty payload
            for i in ('admin', infos_jojo['username']):
                environ = {'REMOTE_USER': i}
                payload = {}
                resp = self.app.put_json('/services_users/?username=jojo',
                                         payload,
                                         extra_environ=environ, status="*")
                self.assertEqual(400, resp.status_int)
                self.assertTrue(b'Nothing to do' in resp.body)
                resp = self.app.put_json(
                    '/services_users/?email=jojo@starplatinum.dom',
                    payload, extra_environ=environ, status="*")
                self.assertEqual(400, resp.status_int)
                self.assertTrue(b'Nothing to do' in resp.body)
                resp = self.app.put_json(
                    '/services_users/?id=%s' % jojo_id,
                    payload, extra_environ=environ, status="*")
                self.assertEqual(400, resp.status_int)
                self.assertTrue(b'Nothing to do' in resp.body)
            # try doing it right
            c = 0
            for i in ('admin', infos_jojo['username']):
                environ = {'REMOTE_USER': i}
                payload = {'full_name': '%i' % c}
                resp = self.app.put_json('/services_users/?username=jojo',
                                         payload,
                                         extra_environ=environ, status="*")
                self.assertEqual(200, resp.status_int)
                self.assertEqual({'updated_fields': payload},
                                 resp.json)
                response = self.app.get('/services_users/?username=jojo',
                                        extra_environ=environ, status="*")
                self.assertEqual(payload['full_name'],
                                 response.json.get('fullname'))
                c += 1
                payload = {'full_name': '%i' % c}
                resp = self.app.put_json(
                    '/services_users/?email=jojo@starplatinum.dom',
                    payload, extra_environ=environ, status="*")
                self.assertEqual(200, resp.status_int)
                self.assertEqual({'updated_fields': payload},
                                 resp.json)
                response = self.app.get('/services_users/?username=jojo',
                                        extra_environ=environ, status="*")
                self.assertEqual(payload['full_name'],
                                 response.json.get('fullname'))
                c += 1
                payload = {'full_name': '%i' % c}
                resp = self.app.put_json(
                    '/services_users/?id=%s' % jojo_id,
                    payload, extra_environ=environ, status="*")
                self.assertEqual(200, resp.status_int)
                self.assertEqual({'updated_fields': payload},
                                 resp.json)
                response = self.app.get('/services_users/?username=jojo',
                                        extra_environ=environ, status="*")
                self.assertEqual(payload['full_name'],
                                 response.json.get('fullname'))
                c += 1


class TestHooksController(FunctionalTest):
    def test_non_existing_hook(self):
        with patch.object(SFGerritProjectManager, 'get_user_groups'):
            environ = {'REMOTE_USER': 'SF_SERVICE_USER'}
            resp = self.app.post_json('/hooks/toto', {'arg1': 1, 'arg2': 2},
                                      extra_environ=environ, status="*")
            self.assertEqual(400, resp.status_int)



#
# Copyright (c) 2016 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


import logging
from managesf import model


logger = logging.getLogger(__name__)


crud = model.SFUserCRUD()


class SFUserMapper:
    def set(self, sf_user_id, service, service_user_id):
        crud.set_service_mapping(sf_user_id, service, service_user_id)

    def get_service_mapping(self, service, sf_user_id):
        """Returns the service uid of user sf_user_id in service."""
        return crud.get_service_mapping(service, sf_user_id)

    def get_user_mapping(self, service, service_user_id):
        """Returns the managesf uid of user service_user_id in service."""
        return crud.get_user_mapping(service, service_user_id)

    def delete(self, sf_user_id, service=None, service_user_id=None):
        return crud.delete_service_mapping(sf_user_id,
                                           service,
                                           service_user_id)


class SFUserManager:
    mapping = SFUserMapper()

    def get(self, id=None, username=None, email=None,
            fullname=None, cauth_id=None):
        return crud.get(id=id, username=username, email=email,
                        fullname=fullname, cauth_id=cauth_id)

    def all(self):
        return self.get()

    def create(self, username, email,
               fullname, cauth_id=None):
        msg = u'Creating user: username=%s, email=%s, full name=%s'
        if cauth_id:
            msg += ', cauth_id=%s' % cauth_id
        logger.info(msg % (username, email, fullname))
        # cauth_id, if set, is our Authoritah (c)
        if cauth_id and cauth_id > 0:
            user = crud.get(cauth_id=cauth_id)
            if user:
                msg = u'Found user %s with cauth_id=%s, updating instead'
                logger.info(msg % (repr(user), cauth_id))
                crud.update(user['id'], username=username,
                            email=email, fullname=fullname)
                return user['id']
        # if not, check if we have a user with the same characteristics:
        user = crud.get(username=username, email=email, fullname=fullname)
        if user:
            if cauth_id and cauth_id != user.get('cauth_id'):
                msg = u'Found user %s, resetting cauth_id instead'
                msg += (' (this is normal if the SSO has been reset or the'
                        ' user has been recreated in the Identity Provider)')
                logger.info(msg % repr(user))
                self.reset_cauth_id(user['id'], cauth_id)
            else:
                msg = u'User %s already exists, doing nothing'
                logger.info(msg % repr(user))
            return user['id']
        return crud.create(username, email, fullname, cauth_id)

    def update(self, id, username=None, email=None, fullname=None,
               idp_sync=None):
        msg = u'Updating user info (id %s):' % id
        if username:
            msg += u' username=%s,' % username
        if email:
            msg += u' email=%s,' % email
        if fullname:
            msg += u' full name=%s' % fullname
        if idp_sync is not None:
            msg += u' idp_sync=%s' % idp_sync
        logger.info(msg)
        return crud.update(id, username=username,
                           email=email, fullname=fullname, idp_sync=idp_sync)

    def reset_cauth_id(self, id, cauth_id):
        msg = u'Updating user info (id %s): cauth_id=%s'
        logger.info(msg % (id, cauth_id))
        return crud.update(id, cauth_id=cauth_id)

    def delete(self, id=None, username=None,
               email=None, fullname=None, cauth_id=None):
        msg = u'Deleting user'
        if id:
            msg += u' (id %s):' % id
        else:
            msg += u':'
        if username:
            msg += u' username=%s,' % username
        if email:
            msg += u' email=%s,' % email
        if fullname:
            msg += u' full name=%s' % fullname
        if cauth_id:
            msg += u', cauth_id=%s' % cauth_id
        logger.info(msg)
        return crud.delete(id, username, email, fullname, cauth_id)
